@@ -13,13 +13,13 @@
 
 package IO::Socket::SSL;
 
-our $VERSION = '2.016';
+our $VERSION = '2.017';
 
 use IO::Socket;
 use Net::SSLeay 1.46;
 use IO::Socket::SSL::PublicSuffix;
 use Exporter ();
-use Errno qw( EWOULDBLOCK ETIMEDOUT EINTR );
+use Errno qw( EWOULDBLOCK EAGAIN ETIMEDOUT EINTR );
 use Carp;
 use strict;
 
@@ -804,7 +804,7 @@ sub _update_peer {
 	my $sockaddr = getpeername( $self );
 	my $af = sockaddr_family($sockaddr);
 	if( CAN_IPV6 && $af == AF_INET6 ) {
-	    my ($host, $port) = getnameinfo($sockaddr,
+	    my (undef, $host, $port) = getnameinfo($sockaddr,
 		NI_NUMERICHOST | NI_NUMERICSERV);
 	    $arg_hash->{PeerAddr} = $host;
 	    $arg_hash->{PeerPort} = $port;
@@ -1111,7 +1111,7 @@ sub readline {
 	    my $rv = $self->sysread($buf,2**16,length($buf));
 	    if ( ! defined $rv ) {
 		next if $! == EINTR;       # retry
-		last if $! == EWOULDBLOCK; # use everything so far
+		last if $! == EWOULDBLOCK || $! == EAGAIN; # use everything so far
 		return;                    # return error
 	    } elsif ( ! $rv ) {
 		last
@@ -1141,7 +1141,7 @@ sub readline {
 	    my $rv = $self->sysread($buf,$size-length($buf),length($buf));
 	    if ( ! defined $rv ) {
 		next if $! == EINTR;       # retry
-		last if $! == EWOULDBLOCK; # use everything so far
+		last if $! == EWOULDBLOCK || $! == EAGAIN; # use everything so far
 		return;                    # return error
 	    } elsif ( ! $rv ) {
 		last
@@ -1494,6 +1494,21 @@ if ( defined &Net::SSLeay::get_peer_cert_chain
 	}
     }
 
+    sub sock_certificate {
+	my ($self,$field) = @_;
+	my $ssl = $self->_get_ssl_object || return;
+	my $cert = Net::SSLeay::get_certificate( $ssl ) || return;
+	if ($field) {
+	    my $sub = $dispatcher{$field} or croak
+		"invalid argument for sock_certificate, valid are: ".join( " ",keys %dispatcher ).
+		"\nMaybe you need to upgrade your Net::SSLeay";
+	    return $sub->($cert);
+	} else {
+	    return $cert
+	}
+    }
+
+
     # known schemes, possible attributes are:
     #  - wildcards_in_alt (0, 'full_label', 'anywhere')
     #  - wildcards_in_cn (0, 'full_label', 'anywhere')
@@ -1719,14 +1734,15 @@ sub get_servername {
 }
 
 sub get_fingerprint_bin {
-    my $cert = shift()->peer_certificate;
-    return Net::SSLeay::X509_digest($cert, $algo2digest->(shift() || 'sha256'));
+    my ($self,$algo,$cert) = @_;
+    $cert ||= $self->peer_certificate;
+    return Net::SSLeay::X509_digest($cert, $algo2digest->($algo || 'sha256'));
 }
 
 sub get_fingerprint {
-    my ($self,$algo) = @_;
+    my ($self,$algo,$cert) = @_;
     $algo ||= 'sha256';
-    my $fp = get_fingerprint_bin($self,$algo) or return;
+    my $fp = get_fingerprint_bin($self,$algo,$cert) or return;
     return $algo.'$'.unpack('H*',$fp);
 }
 
@@ -2082,24 +2098,28 @@ sub new {
 	 defined( my $file = $arg_hash->{$_} ) or next;
 	for my $f (ref($file) eq 'HASH' ? values(%$file):$file ) {
 	    die "$_ $f does not exist" if ! -f $f;
-	    die "$_ $f is not accessible" if ! -r _;
 	}
     }
 
-    my $verify_mode = $arg_hash->{SSL_verify_mode};
+    my $verify_mode = $arg_hash->{SSL_verify_mode} || 0;
     if ( $verify_mode != Net::SSLeay::VERIFY_NONE()) {
 	if ( defined( my $f = $arg_hash->{SSL_ca_file} )) {
 	    if ( ! ref($f) || $$f ) {
 		die "SSL_ca_file $f does not exist" if ! -f $f;
-		die "SSL_ca_file $f is not accessible" if ! -r _;
+		die "SSL_ca_file $f is not accessible: $!" 
+		    if ! open(my $fh,'<',$f);
 	    }
 	}
 	if ( defined( my $d = $arg_hash->{SSL_ca_path} )) {
 	    if ( ! ref($d) || $$d ) {
 		die "SSL_ca_path $d does not exist" if ! -d $d;
-		die "SSL_ca_path $d is not accessible" if ! -r _;
+		die "SSL_ca_path $d is not accessible: $!" 
+		    if ! opendir(my $dh,$d);
 	    }
 	}
+    } elsif ( $verify_mode ne '0' ) {
+	# some users use the string 'SSL_VERIFY_PEER' instead of the constant
+	die "SSL_verify_mode must be a number and not a string";
     }
 
     my $self = bless {},$class;
@@ -2141,24 +2161,6 @@ sub new {
 	    # verify name
 	    my $rv = IO::Socket::SSL::verify_hostname_of_cert(
 		$host,$cert,$vcn_scheme,$vcn_publicsuffix );
-	    if ( ! $rv && ! $vcn_scheme ) {
-		# For now we use the default hostname verification if none
-		# was specified and complain loudly but return ok if it does
-		# not match. In the future we will enforce checks and users
-		# should better specify and explicite verification scheme.
-		warn <<WARN;
-
-The verification of cert '$certname'
-failed against the host '$host' with the default verification scheme.
-
-   THIS MIGHT BE A MAN-IN-THE-MIDDLE ATTACK !!!!
-
-To stop this warning you might need to set SSL_verifycn_name to
-the name of the host you expect in the certificate.
-
-WARN
-		return $ok;
-	    }
 	    if ( ! $rv ) {
 		IO::Socket::SSL->_internal_error(
 		    "hostname verification failed",5);
@@ -2502,7 +2504,7 @@ WARN
 		Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_subject_name($cert));
 	    $error &&= Net::SSLeay::ERR_error_string($error);
 	}
-	$DEBUG>=3 && DEBUG( "ok=$ok cert=$cert" );
+	$DEBUG>=3 && DEBUG( "ok=$ok [$depth] $certname" );
 	$ok = $verify_cb->($ok,$ctx_store,$certname,$error,$cert,$depth) if $verify_cb;
 	$ok = $verify_fingerprint->($ok,$cert,$depth) if $verify_fingerprint && $cert;
 	return $ok;
@@ -2609,9 +2611,11 @@ WARN
     }
 
     if ( my $cl = $arg_hash->{SSL_cipher_list} ) {
-	for (values %ctx) {
-	    Net::SSLeay::CTX_set_cipher_list($_, $cl ) || return 
-		IO::Socket::SSL->error("Failed to set SSL cipher list");
+	for (keys %ctx) {
+	    Net::SSLeay::CTX_set_cipher_list($ctx{$_}, ref($cl) 
+		? $cl->{$_} || $cl->{''} || $DEFAULT_SSL_ARGS{SSL_cipher_list} || next 
+		: $cl
+	    ) || return IO::Socket::SSL->error("Failed to set SSL cipher list");
 	}
     }
 
